!=======================================================================
! PC/HD IDRIS le 16-02-2001
!
!      ********************************************************
!      **  Cours IDRIS Fortran 90 : corrige de l'exercice 10 **
!      ********************************************************
!
! Voici la sortie du programme avec le fichier exo10.data
! =======================================================
!
!             Matrice : 
!
! -2.000000000 2.000000000 2.000000000 2.000000000
! -3.000000000 3.000000000 2.000000000 2.000000000
! -2.000000000 0.000000000 4.000000000 2.000000000
! -1.000000000 0.000000000 0.000000000 5.000000000
!
!
!            Valeurs propres : 
!
! Valeur propre N. 1 : ( 4.00E+00  ,  0.00E+00)
! Valeur propre N. 2 : ( 3.00E+00  ,  0.00E+00)
! Valeur propre N. 3 : ( 2.00E+00  ,  0.00E+00)
! Valeur propre N. 4 : ( 1.00E+00  ,  0.00E+00)
!=======================================================================


module matrix
  use f95_lapack, only: GEEV => LA_GEEV
  implicit none
  integer          :: nb_lignes=0, nb_colonnes=0
  integer, private :: err

  type OBJ_MAT
    private
    integer                           :: n=0, m=0
    real, dimension(:,:), allocatable :: mat
  end type OBJ_MAT

  private  :: add, trans, taille_mat, valorisation, affect
  private  :: val_propres
!--------------------------------
  interface operator(+)
     module procedure add
  end interface
!--------------------------------
  interface operator(.tr.)
    module procedure trans
  end interface
!--------------------------------
  interface operator(.vp.)
     module procedure val_propres
  end interface
!--------------------------------
  interface assignment(=)
    module procedure taille_mat, valorisation, affect
  end interface
!--------------------------------
contains
!--------------------------------
  subroutine valorisation(a,t)
    type(OBJ_MAT),      intent(inout) :: a
    real, dimension(:), intent(in)    :: t

    if ( nb_lignes <=0 .or. nb_colonnes <= 0 ) then
      print *, 'Valorisation impossible.'
      print *, 'Veuillez valoriser le profil &
               &(nb_lignes,nb_colonnes) &
               &avec des valeurs acceptables.'
    else
      if (.not.allocated(a%mat)) then
        allocate( a%mat(nb_lignes,nb_colonnes), stat=err )
        if (err /= 0) then
          print *,"Valorisation : problme mmoire."
          stop 4
        endif
        a%n = nb_lignes
        a%m = nb_colonnes
      endif
      call affectation
    end if

  contains
    subroutine affectation
      real, dimension(size(a%mat)-size(t)) :: padding

      padding = 0.
      a%mat = reshape( source = t,            &
                       shape  = shape(a%mat), &
                       pad    = padding )
    end subroutine affectation
  end subroutine valorisation
!--------------------------------
  subroutine poubelle(a)
    type(OBJ_MAT), intent(inout) :: a

    if (allocated(a%mat)) then
      a%n = 0; a%m = 0
      deallocate(a%mat)
    endif
  end subroutine poubelle
!--------------------------------
  function add(a,b)
    type(OBJ_MAT), intent(in) :: a,b
    type(OBJ_MAT)             :: add

    allocate(add%mat(a%n,a%m),stat=err)
    if (err /= 0) then
      print *,"Somme : problme mmoire."
      stop 8
    endif
    add%n = a%n
    add%m = a%m
    add%mat = a%mat + b%mat
  end function add
!--------------------------------
  function trans(a)
    type(OBJ_MAT), intent(in) :: a
    type(OBJ_MAT)             :: trans

    allocate(trans%mat(a%m,a%n),stat=err)
    if (err /= 0) then
      print *,"Transpos : problme mmoire."
      stop 16
    endif
    trans%n = a%m; trans%m = a%n
    trans%mat = transpose(a%mat)
  end function trans
!--------------------------------
  function val_propres(a)
    type(OBJ_MAT), intent(in)     :: a
    complex, dimension(nb_lignes) :: val_propres
    real,    dimension(nb_lignes)        :: Wr, Wi
    real,    dimension(:,:), allocatable :: matrice
 
    if (allocated(a%mat)) then
      matrice = a%mat
      call GEEV(matrice, Wr, Wi)
      val_propres = cmplx(Wr, Wi)
    else
      print *, "Objet non existant"
    end if
  end function val_propres
!--------------------------------
  subroutine taille_mat(i,a)
    integer,       intent(out) :: i
    type(OBJ_MAT), intent(in)  :: a

    i = a%n*a%m
  end subroutine taille_mat
!--------------------------------
  subroutine affect(a,b)
    type(OBJ_MAT), intent(inout) :: a
    type(OBJ_MAT), intent(in)    :: b

    if (.not.allocated(b%mat)) &
      stop "Erreur ==> membre de droite de l'affectation non &
          &initialis"
    if (allocated(a%mat) .and. &
        any(shape(a%mat) /= shape(b%mat))) &
      stop "Erreur ==> affectation entre matrices non conformantes"
    if (.not.allocated(a%mat)) then
      allocate(a%mat(b%n,b%m), stat=err)
      if (err /= 0) &
        stop "Erreur ==> allocation membre de gauche mal passe"
    end if
    a%n = b%n ; a%m = b%m
    a%mat = b%mat
  end subroutine affect      
!--------------------------------
  subroutine imp(a)
    type(OBJ_MAT), intent(in)  :: a
    integer(kind=2)            :: i

    if ( .not.allocated(a%mat) ) then
      print *,"Impression : objet inexistant."
      stop 32
    end if
    print '(//, a, /)', "             Matrice : "
    do i=1,size(a%mat,1)
      print '(*(f6.2))',a%mat(i,:) ! Format : Fortran 2008
    enddo
  end subroutine imp
!--------------------------------
   subroutine imp_vp(vec)
     complex, dimension(:) :: vec
     integer               :: i

     print '(//, a, /)', "            Valeurs propres : "
     do i=1,size(vec)
       print '("Valeur propre N.", i2, " : (", 1pe9.2, "  , ", 1pe9.2, ")")', &
              i, real(vec(i)), aimag(vec(i))
     end do
   end subroutine imp_vp
end module matrix

program exo11
  use matrix
  type(OBJ_MAT)                      :: u
  real,    dimension(:), allocatable :: val_init
  complex, dimension(:), allocatable :: val_pr
  open(unit=10, file='exo11.data', form='formatted', action='read')
  read(10, advance='no', fmt='(i1)') nb_lignes
  nb_colonnes = nb_lignes
  allocate(val_init(nb_lignes*nb_colonnes))
  read(10, *) val_init 
  close(10)
  u = val_init
  call imp(u)
! ---------------
  val_pr = .vp. u
! ---------------
  call imp_vp(val_pr)
  call poubelle(u)
end program exo11
