module music
 implicit none
 integer, parameter  :: nb_enr=30
 integer             :: nb_mus
!---------------------------------
 type musicien
   private
   character(len=16) :: prenom
   character(len=21) :: nom
   integer           :: annee_naiss,annee_mort
 end type musicien
!---------------------------------
 type , private :: ptr_musicien
   type(musicien), pointer :: ptr
 end type ptr_musicien
!---------------------------------
 type(musicien), dimension(nb_enr), target :: tab_mus
 type(ptr_musicien), dimension(:), pointer :: tab_ptr_musicien
 private                                   :: tab_ptr_musicien 
!---------------------------------
 interface operator(<)
   module procedure longevite
 end interface
!---------------------------------
 interface operator(.MortLePlusJeune.)
   module procedure mort_le_plus_jeune
 end interface
!---------------------------------
 private :: operator(<), longevite, mort_le_plus_jeune
contains
!---------------------------------
 subroutine init
   integer :: eof,i
!
! Valorisation du tableau de musiciens
!
   open(1,file="musiciens",action="read",status="old")
   nb_mus = 0
   do
     read(1,'(a,1x,a,2(1x,i4))',iostat=eof) tab_mus(nb_mus+1)
     if (eof /= 0) exit
     nb_mus = nb_mus + 1
   enddo
   close(1)
!
! On alloue le tableau de pointeurs dont le nombre d'lments
! correspond au nombre de musiciens.
!
   allocate(tab_ptr_musicien(nb_mus))
!
! Chaque lment du tableau de pointeurs alloue prcdemment va
! tre mis en relation avec l'lment correspondant du tableau de
! musiciens.
! N.B.: Chaque lment du tableau de musiciens (tab_mus)  l'attribut
! target implicitement car cet attribut a t spcifi pour le tableau
! lui-mme.
!
   do i=1,nb_mus
     tab_ptr_musicien(i)%ptr => tab_mus(i)
   enddo
   print *,'---- Liste des musiciens ----'
   print *
   write(*,'((5x,a,1x,a,2(1x,i4)))') (tab_mus(i),i=1,nb_mus)
 end subroutine init
!---------------------------------
 subroutine tri(critere)
!
! Cette procdure va trier la liste des musiciens soit par ordre
! alphabtique des noms soit par ordre chronologique en fonction
! du paramtre critre spcifi.
! Ce tri s'effectuera par l'intermdiaire du tableau de pointeurs
! tab_ptr_musicien.
!
  ! Dclaration de l'argument
  character(len=*), intent(in) :: critere
  ! Dclarations locales
  character(len=*), parameter, dimension(2) :: labels = &
          [ "alphabetique ", "chronologique" ]
  character(len=len(labels))                :: label_tri
  logical                                   :: expr
  integer                                   :: i, j

  do i=nb_mus-1,1,-1
    do j=1,i
      select case(critere)
        case("nom")
          label_tri = labels(1)
          expr = lgt(tab_ptr_musicien(j)%ptr%nom, &
                     tab_ptr_musicien(j+1)%ptr%nom)
        case("annee")
          label_tri = labels(2)
          expr = tab_ptr_musicien(j)%ptr%annee_naiss > &
                 tab_ptr_musicien(j+1)%ptr%annee_naiss
        case default
      end select
      if (expr) &
        !-------Permutation des deux associations-----------
        tab_ptr_musicien(j:j+1) = tab_ptr_musicien(j+1:j:-1)
        !---------------------------------------------------
    enddo
  enddo
  print '(/, a, a, a, /)', '---- Liste ', trim(label_tri), ' des musiciens ----'
  write(*,'((5x,a,1x,a,2(1x,i4)))') (tab_ptr_musicien(i)%ptr,i=1,nb_mus)
 end subroutine tri
!---------------------------------
 function longevite(mus1,mus2)
!
! Fonction permettant de surcharger l'oprateur <
! afin de pouvoir specifier des oprandes de type musicien.
!
   type(musicien), intent(in) :: mus1,mus2
   logical                    :: longevite
   integer                    :: duree_de_vie_mus1,duree_de_vie_mus2
   duree_de_vie_mus1 = mus1%annee_mort - mus1%annee_naiss
   duree_de_vie_mus2 = mus2%annee_mort - mus2%annee_naiss
   longevite = duree_de_vie_mus1 < duree_de_vie_mus2
 end function longevite
!---------------------------------
 function mort_le_plus_jeune(tab_mus)
!
! procedure de dfinition de l'oprateur .MortLePlusJeune.
!
   type(musicien), dimension(:), intent(in) :: tab_mus
   type(musicien)                           :: mort_le_plus_jeune
   integer i

   mort_le_plus_jeune = tab_mus(1)
   do i=2,nb_mus
!
! N.B.: ici l'utilisation de l'oprateur < provoque
! ---   l'appel  la fonction longevite.
!       En fait : tab_mus(i) < mus  <=> longevite(tab_mus(i),mus)

     if (tab_mus(i) < mort_le_plus_jeune) &
                    mort_le_plus_jeune = tab_mus(i)
   enddo
 end function mort_le_plus_jeune
!---------------------------------
 function nom(mus)
!
! Fonction renvoyant les nom et prenom du musicien
! pass en argument.
!
   type(musicien), intent(in) :: mus
   character(len=38)          :: nom
!  write(nom, '(a,1x,a)')mus%prenom,mus%nom
   nom = trim(mus%prenom)//' '//mus%nom
 end function nom
end module music
