module music
 implicit none
 type musicien
   private
   character(len=16)       :: prenom
   character(len=21)       :: nom
   integer                 :: annee_naiss,annee_mort
   type(musicien), pointer :: ptr_precedent, ptr_suivant
 end type musicien
!
 type(musicien), pointer   :: debut
!
 interface operator(<)
   module procedure longevite
 end interface
!
 interface operator(.MortLePlusJeune.)
   module procedure mort_le_plus_jeune
 end interface
!------------------------------------------------------------
 private :: operator(<), longevite, mort_le_plus_jeune, liste
contains
  subroutine init
    type(musicien)          :: mus
    type(musicien), pointer :: ptr_precedent, ptr_courant
    integer                 :: eof

    nullify(debut)
    nullify(mus%ptr_precedent)
    nullify(mus%ptr_suivant)
    open(1,file="musiciens",action="read",status="old")
    do
      read(1,'(a,1x,a,2(1x,i4))',iostat=eof)mus%prenom,      &
                                            mus%nom,         &
                                            mus%annee_naiss, &
                                            mus%annee_mort
      if (eof /= 0) exit
      allocate(ptr_courant)
      ptr_courant = mus
      if (.not.associated(debut)) then
        debut => ptr_courant
      else
        ptr_precedent%ptr_suivant => ptr_courant
        ptr_courant%ptr_precedent => ptr_precedent
      endif
      ptr_precedent => ptr_courant
    enddo
    close(1)
  
    print *
    print *,'---- Liste des musiciens ----'
    print *
    call liste
  end subroutine init
!
  subroutine tri(critere)
  !
  ! Cette procdure va trier la liste des musiciens soit par ordre
  ! alphabtique des noms soit par ordre chronologique en fonction
  ! du paramtre critre specifi.
  !
    ! Dclaration de l'argument
    character(len=*), intent(in) :: critere
    ! Dclarations locales
    character(len=*), parameter, dimension(2) :: labels = &
            [ "alphabetique ", "chronologique" ]
    character(len=len(labels))                :: label_tri
    type(musicien), pointer                   :: ptr_courant, ptr
    logical                                   :: tri_termine, expr
  
    do
      tri_termine = .true.
      ptr_courant => debut
      do
        if (.not.associated(ptr_courant%ptr_suivant)) exit
        select case(critere)
          case("nom")
            label_tri = labels(1)
            expr = ptr_courant%nom > ptr_courant%ptr_suivant%nom
          case("annee")
            label_tri = labels(2)
            expr = ptr_courant%annee_naiss > &
                   ptr_courant%ptr_suivant%annee_naiss
          case default
        end select
        if (expr) then
          allocate(ptr)
          ptr = ptr_courant%ptr_suivant
          call insere(ptr_courant, ptr)
          call suppression(ptr_courant%ptr_suivant)
          tri_termine = .false.
          cycle
        end if
        if (associated(ptr_courant%ptr_suivant)) &
                       ptr_courant   => ptr_courant%ptr_suivant
      end do
      if (tri_termine) exit
    end do

    print *
    print *,'---- Liste ', trim(label_tri), ' des musiciens ----'
    print *
    call liste
  end subroutine tri
!
  subroutine insere(ptr_courant, ptr)
    type(musicien), pointer :: ptr_courant, ptr

    if (associated(ptr_courant, debut)) then
      debut => ptr
    else
      ptr_courant%ptr_precedent%ptr_suivant => ptr
    end if
    ptr%ptr_suivant => ptr_courant
    ptr%ptr_precedent => ptr_courant%ptr_precedent  
    ptr_courant%ptr_precedent => ptr
  end subroutine insere
!
  subroutine suppression(ptr)
    type(musicien), pointer :: ptr
    type(musicien), pointer :: temp

    temp => ptr
    ptr  => ptr%ptr_suivant
    if (associated(temp%ptr_suivant)) &
      temp%ptr_suivant%ptr_precedent => temp%ptr_precedent
    deallocate(temp)
  end subroutine suppression
!
!---------------------------------
 function longevite(mus1,mus2)
!
! Fonction permettant de surcharger l'oprateur <
! afin de pouvoir spcifier des oprandes de type musicien.
!
   type(musicien), intent(in) :: mus1,mus2
   logical                    :: longevite
   integer                    :: duree_de_vie_mus1,duree_de_vie_mus2
   duree_de_vie_mus1 = mus1%annee_mort - mus1%annee_naiss
   duree_de_vie_mus2 = mus2%annee_mort - mus2%annee_naiss
   longevite = duree_de_vie_mus1 < duree_de_vie_mus2
 end function longevite
!---------------------------------
 function mort_le_plus_jeune(debut)
   type(musicien), intent(in) :: debut
   type(musicien)             :: mort_le_plus_jeune
   type(musicien), pointer    :: p_mus

   mort_le_plus_jeune = debut; p_mus => debut%ptr_suivant
   do while(associated(p_mus))
!
! N.B.: ici l'utilisation de l'oprateur < provoque
! ---   l'appel  la fonction longevite.
!       En fait : p_mus < mort_le_plus_jeune  <=>
!                 longevite(p_mus, mort_le_plus_jeune)

     if (p_mus < mort_le_plus_jeune) mort_le_plus_jeune = p_mus
     p_mus => p_mus%ptr_suivant
   enddo
 end function mort_le_plus_jeune
!---------------------------------
 function nom(mus)
!
! Fonction renvoyant les nom et prenom du musicien
! pass en argument.
!
   type(musicien), intent(in) :: mus
   character(len=38)          :: nom
!  write(nom, '(a,1x,a)')mus%prenom,mus%nom
   nom = trim(mus%prenom)//' '//mus%nom
 end function nom
!---------------------------------
  subroutine liste
    type(musicien), pointer :: ptr_courant

    ptr_courant => debut
    if (.not.associated(debut)) then
      print *,"Il n'existe aucun musicien!"
      stop  8
    end if
    do
     write(*,'((5x,a,1x,a,2(1x,i4)))')ptr_courant%prenom,      &
                                      ptr_courant%nom,         &
                                      ptr_courant%annee_naiss, &
                                      ptr_courant%annee_mort
     if (.not.associated(ptr_courant%ptr_suivant)) exit
     ptr_courant => ptr_courant%ptr_suivant
    end do
    print *
    print *,"liste inversee"
    print *,"--------------"
    print *
    do
     write(*,'((5x,a,1x,a,2(1x,i4)))')ptr_courant%prenom,      &
                                      ptr_courant%nom,         &
                                      ptr_courant%annee_naiss, &
                                      ptr_courant%annee_mort
     if (associated(ptr_courant, debut)) exit
     ptr_courant => ptr_courant%ptr_precedent
    end do
  end subroutine liste
end module music
!
program exo13
 use music
 type(musicien) :: mus_mort_le_plus_jeune
 call init
 call tri(critere="nom")
 call tri(critere="annee")
 mus_mort_le_plus_jeune = .MortLePlusJeune.debut
 print *
 print *,"Le musicien mort le plus jeune est : ", &
          nom(mus_mort_le_plus_jeune)
end program exo13
