module music
 implicit none
 type musicien
   private
   character(len=16)       :: prenom
   character(len=21)       :: nom
   integer                 :: annee_naiss,annee_mort
   type(musicien), pointer :: ptr_precedent => NULL()
   type(musicien), pointer :: ptr_suivant   => NULL()
 end type musicien
!
 type, private :: ptr_mus
   type(musicien), pointer :: ptr
 end type ptr_mus
!
 type(musicien), pointer :: debut => NULL()
 type(ptr_mus), dimension(:), allocatable, private :: tab_ptr_musicien
 integer, private :: nb_mus
!
 interface operator(<)
   module procedure longevite
 end interface
!
 interface operator(.MortLePlusJeune.)
   module procedure mort_le_plus_jeune
 end interface

 private :: operator(<), longevite, mort_le_plus_jeune, liste
!------------------------------------------------------------
contains
  subroutine init
    type(musicien)          :: mus
    type(musicien), pointer :: ptr_precedent, ptr_courant
    integer                 :: eof

    open(1,file="musiciens",action="read",status="old")
    nb_mus = 0
    do
      read(1,'(a16,1x,a21,2(1x,i4))',iostat=eof) mus%prenom,      &
                                                 mus%nom,         &
                                                 mus%annee_naiss, &
                                                 mus%annee_mort
      if (eof /= 0) exit
      allocate(ptr_courant)
      ptr_courant = mus
      if (.not.associated(debut)) then
        debut => ptr_courant
      else
        ptr_precedent%ptr_suivant => ptr_courant
        ptr_courant%ptr_precedent => ptr_precedent
      endif
      ptr_precedent => ptr_courant
      nb_mus = nb_mus + 1
    enddo
    close(1)
  
    call association

    print *
    print *,'---- Liste des musiciens ----'
    print *
    call liste

  contains
    subroutine association
      integer i

      allocate( tab_ptr_musicien(nb_mus) )
      ptr_courant => debut
      do i=1,nb_mus
        tab_ptr_musicien(i)%ptr => ptr_courant
        ptr_courant => ptr_courant%ptr_suivant
      end do
    end subroutine association
  end subroutine init
!
  subroutine tri(critere)
    !
    ! Cette procdure va trier la liste des musiciens soit par ordre
    ! alphabtique des noms soit par ordre chronologique en fonction
    ! du paramtre critre specifi.
    !
    ! Dclaration de l'argument
    character(len=*), intent(in) :: critere
    ! Dclarations locales
    character(len=*), parameter, dimension(2) :: labels = &
            [ "alphabetique ", "chronologique" ]
    character(len=len(labels))                :: label_tri
    logical                                   :: expr
    integer                                   :: i, j
  
    do i=nb_mus-1,1,-1
      do j=1,i
        select case(critere)
          case("nom")
            label_tri = labels(1)
            expr = lgt(tab_ptr_musicien(j)%ptr%nom, &
                       tab_ptr_musicien(j+1)%ptr%nom)
          case("annee")
            label_tri = labels(2)
            expr = tab_ptr_musicien(j)%ptr%annee_naiss > &
                   tab_ptr_musicien(j+1)%ptr%annee_naiss
        end select
        if (expr) &
          !-------Permutation des deux associations-----------
          tab_ptr_musicien(j:j+1) = tab_ptr_musicien(j+1:j:-1)
          !---------------------------------------------------
      enddo
    enddo

    print *
    print *,'---- Liste ', trim(label_tri), ' des musiciens ----'
    print *
    call liste
  end subroutine tri
!
!---------------------------------
  function longevite(mus1,mus2)
!
! Fonction permettant de surcharger l'oprateur <
! afin de pouvoir spcifier des oprandes de type musicien.
!
    type(musicien), intent(in) :: mus1,mus2
    logical                    :: longevite
    integer                    :: duree_de_vie_mus1,duree_de_vie_mus2
    duree_de_vie_mus1 = mus1%annee_mort - mus1%annee_naiss
    duree_de_vie_mus2 = mus2%annee_mort - mus2%annee_naiss
    longevite = duree_de_vie_mus1 < duree_de_vie_mus2
  end function longevite
!---------------------------------
  function mort_le_plus_jeune(debut)
    type(musicien), intent(in) :: debut
    type(musicien)             :: mort_le_plus_jeune
    type(musicien), pointer    :: p_mus
 
    mort_le_plus_jeune = debut; p_mus => debut%ptr_suivant
    do while(associated(p_mus))
!
! N.B.: ici l'utilisation de l'oprateur < provoque
! ---   l'appel  la fonction longevite.
!       En fait : p_mus < mort_le_plus_jeune  <=>
!                 longevite(p_mus, mort_le_plus_jeune)
 
      if (p_mus < mort_le_plus_jeune) mort_le_plus_jeune = p_mus
      p_mus => p_mus%ptr_suivant
    enddo
  end function mort_le_plus_jeune
!---------------------------------
  function nom(mus)
!
! Fonction renvoyant les nom et prenom du musicien
! pass en argument.
!
    type(musicien), intent(in) :: mus
    character(len=38)          :: nom
    nom = trim(mus%prenom)//' '//mus%nom
  end function nom
!---------------------------------
  subroutine liste
    integer i

    if ( nb_mus == 0 ) then
      print *,"Il n'existe aucun musicien!"
      stop  8
    end if
    do i=1,nb_mus
      write(*,'((5x,a16,1x,a21,2(1x,i4)))') &
                               tab_ptr_musicien(i)%ptr%prenom,      &
                               tab_ptr_musicien(i)%ptr%nom,         &
                               tab_ptr_musicien(i)%ptr%annee_naiss, &
                               tab_ptr_musicien(i)%ptr%annee_mort
    end do
    print *
    print *,"liste inversee"
    print *,"--------------"
    print *
    do i=nb_mus,1,-1
      write(*,'((5x,a16,1x,a21,2(1x,i4)))') &
                               tab_ptr_musicien(i)%ptr%prenom,      &
                               tab_ptr_musicien(i)%ptr%nom,         &
                               tab_ptr_musicien(i)%ptr%annee_naiss, &
                               tab_ptr_musicien(i)%ptr%annee_mort
    end do
  end subroutine liste
end module music
!
program exo13
 use music
 type(musicien) :: mus_mort_le_plus_jeune
 call init
 call tri(critere="nom")
 call tri(critere="annee")
 mus_mort_le_plus_jeune = .MortLePlusJeune.debut
 print *
 print *,"Le musicien mort le plus jeune est : ", &
          nom(mus_mort_le_plus_jeune)
end program exo13
